from abc import ABC

from .stage_interface import StageInterface as ImplementsStageInterface
from ..partials.with_base.types import T_in, T_out
from ..partials.with_base.stage import (
    Stage as WithStageBaseConcern,
)
from ..partials.with_stage_as_callable.stage_mixin import (
    StageMixin as WithStageAsCallableConcern,
)


class Stage(
    WithStageAsCallableConcern[T_in, T_out],
    WithStageBaseConcern,
    ImplementsStageInterface,
    ABC,
):
    pass



from abc import ABC
from typing import Any

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.types import T_in, T_out


class PipelineMixin(
    ImplementsPipelineInterface[T_in, T_out],
    ABC,
):
    async def __call__(self, payload: T_in, *args: Any, **kwds: Any) -> T_out:
        return await self.process(payload, *args, **kwds)



from abc import abstractmethod
from typing import Protocol, Self
from .item import TItem
from .act_as_list_interface import ActAsListInterface


class ActAsList(
    ActAsListInterface[TItem],
    Protocol[TItem],
):
    @abstractmethod
    def _get_items(self) -> list[TItem]:
        pass

    @abstractmethod
    def _set_items(self, items: list[TItem]) -> Self:
        pass

    def _add_item(self, item: TItem) -> Self:
        self._get_items().append(item)

        return self



from typing import Protocol

from ..partials.with_base.types import T_in, T_out
from ..partials.with_base.stage_callable import (
    StageCallable as WithStageCallableBase,
)


class StageCallable(
    WithStageCallableBase[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass



from abc import ABC
from typing import Any, Optional, Self

from ..with_base.types import T_in, T_out
from .processor_interface import ProcessorInterface
from .pipeline_factory_interface_mixin import (
    PipelineFactoryInterfaceMixin as PipelineFactoryInterface,
)


class PipelineFactoryMixin(
    PipelineFactoryInterface[T_in, T_out],
    ABC,
):
    processor_class: Optional[type[ProcessorInterface[T_in, T_out]]] = None
    processor: Optional[ProcessorInterface[T_in, T_out]] = None

    def __init__(
        self,
        processor_class: Optional[type[ProcessorInterface]] = None,
        processor: Optional[ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not self.processor_class:
            self.processor_class = processor_class or None

        if not self.processor:
            self.processor = processor or None

    def with_processor(self, processor: ProcessorInterface) -> Self:
        self.processor = processor

        return self

    def with_processor_class(
        self, processor_class: type[ProcessorInterface]
    ) -> Self:
        self.processor_class = processor_class

        return self

    # ActAsFactory
    def _definition(self) -> dict[str, Any]:
        return {
            **super()._definition(),  # type: ignore
            "processor_class": self.processor_class,
            "processor": self.processor,
        }



from ...classes.pipeline import Pipeline
from ...classes.types import T_in, T_out
from .chained_processor import ChainedProcessor


class ChainedPipeline(Pipeline[T_in, T_out]):
    processor_class = ChainedProcessor



from typing import cast

from ...classes.command import Command
from ...partials.with_base.types import T_in, T_out


class ChainedCommand(Command[T_in, T_out]):
    async def execute(self) -> T_out:
        for stage in self.stages:
            self.payload = await self._call_stage(
                payload=self.payload,
                stage=stage,
                *self.extra_args,
                **self.extra_kwds,
            )

        return cast(T_out, self.payload)



from typing import Any, Awaitable, Protocol

from .types import T_in, T_out


class StageCallable(Protocol[T_in, T_out]):
    def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out | Awaitable[T_out]: ...


StageCallableType = StageCallable[Any, Any]



import copy
from typing import Protocol, Self

from ..act_as_list_interface import ActAsListInterface
from ..item import TItem


class HasListImmutability(
    ActAsListInterface[TItem],
    Protocol[TItem],
):
    def _add_item(self, item: TItem) -> Self:
        cloned = copy.deepcopy(self)

        cloned._get_items().append(item)

        return cloned



from .processor import Processor
from .pipeline_interface import (
    PipelineInterface as ImplementsPipelineInterface,
)
from ..processors.chained_processor.chained_processor import ChainedProcessor
from ..partials.with_base.types import T_in, T_out
from ..partials.with_base.pipeline import (
    Pipeline as WithPipelineBaseConcern,
)
from ..partials.with_pipeline_as_list.pipeline_mixin import (
    PipelineMixin as WithPipelineAsListConcern,
)
from ..partials.with_pipeline_declared_stages.pipeline_mixin import (
    PipelineMixin as WithPipelineDeclaredStagesConcern,
)
from ..partials.with_pipeline_processor.pipeline_mixin import (
    PipelineMixin as WithPipelineProcessorConcern,
)
from ..partials.with_pipeline_as_stage.pipeline_mixin import (
    PipelineMixin as WithPipelineAsStageConcern,
)
from ..partials.with_pipeline_as_immutable.pipeline_mixin import (
    PipelineMixin as WithPipelineAsImmutableConcern,
)
from ..partials.with_processor_commands.pipeline_mixin import (
    PipelineMixin as WithProcessorCommandsConcern,
)


class Pipeline(
    WithPipelineAsImmutableConcern,
    WithPipelineAsStageConcern[T_in, T_out],
    WithProcessorCommandsConcern[T_in, T_out],
    WithPipelineProcessorConcern[T_in, T_out],
    WithPipelineDeclaredStagesConcern,
    WithPipelineAsListConcern,
    WithPipelineBaseConcern,
    ImplementsPipelineInterface[T_in, T_out],
):
    base_processor_class = Processor
    processor_class = ChainedProcessor



from typing import Any

from ...classes.pipeline import Pipeline
from ...classes.types import T_in, T_out
from .interruptible_processor import InterruptibleProcessor, CheckCallable


class InterruptiblePipeline(Pipeline[T_in, T_out]):
    def __init__(self, check: CheckCallable, *args: Any, **kwargs: Any):
        processor = InterruptibleProcessor[T_in, T_out](check)

        super().__init__(processor=processor, *args, **kwargs)



import pytest
from thecodecrate_pipeline import (
    Pipeline,
    ChainedProcessor,
    Command,
    CommandProcessingStrategy,
    Processor,
    ProcessableAsCommand,
    ProcessorProcessingStrategy,
)


@pytest.mark.asyncio
async def test_custom_command_class():
    class MyCommand(Command[int, int]):
        async def execute(self) -> int:
            for stage in self.stages:
                self.payload = await self._call_stage(
                    payload=self.payload,
                    stage=stage,
                    *self.extra_args,
                    **self.extra_kwds,
                )

            return self.payload * 10

    class MyProcessor(Processor[int, int]):
        command_class = MyCommand

    class MyPipeline(Pipeline[int, int]):
        processor_class = MyProcessor

    pipeline = (
        MyPipeline()
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
        .pipe(lambda payload: payload * 3)
    )

    assert await pipeline.process(5) == 360


@pytest.mark.asyncio
async def test_trait_processable_as_command():
    class MyCommand(Command[int, int]):
        async def execute(self) -> int:
            for stage in self.stages:
                self.payload = await self._call_stage(
                    payload=self.payload,
                    stage=stage,
                )

            return self.payload * 5

    class MyProcessor(Processor[int, int]):
        command_class = MyCommand

    class MyPipeline(
        Pipeline[int, int],
        ProcessableAsCommand,
    ):
        processor_class = MyProcessor

    pipeline = (
        MyPipeline()
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
        .pipe(lambda payload: payload * 3)
    )

    assert await pipeline.process(5) == 180


@pytest.mark.asyncio
async def test_set_processor_and_strategy_on_pipeline_constructor():
    class TimesTenCommand(Command[int, int]):
        async def execute(self) -> int:
            for stage in self.stages:
                self.payload = await self._call_stage(
                    payload=self.payload,
                    stage=stage,
                )

            return self.payload * 5

    class MyProcessor(ChainedProcessor[int, int]):
        command_class = TimesTenCommand

    pipeline = (
        Pipeline(
            processor=MyProcessor(
                processing_strategy=CommandProcessingStrategy
            )
        )
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
        .pipe(lambda payload: payload * 3)
    )
    assert await pipeline.process(5) == 180

    pipeline = (
        Pipeline(
            processor=MyProcessor(
                processing_strategy=ProcessorProcessingStrategy
            )
        )
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
        .pipe(lambda payload: payload * 3)
    )
    assert await pipeline.process(5) == 36


@pytest.mark.asyncio
async def test_chained_processor_as_command():
    processor = ChainedProcessor[int](
        processing_strategy=CommandProcessingStrategy
    )

    result = await processor.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
    )
    assert result == 12

    processor = processor.with_processing_strategy_class(
        ProcessorProcessingStrategy
    )

    result = await processor.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
        ],
    )
    assert result == 6


@pytest.mark.asyncio
async def test_immutability_of_process_with_strategy():
    # define a command class
    class TimesTenCommand(Command[int, int]):
        async def execute(self) -> int:
            for stage in self.stages:
                self.payload = await self._call_stage(
                    payload=self.payload,
                    stage=stage,
                )

            return self.payload * 10

    class MyProcessor(ChainedProcessor[int, int]):
        command_class = TimesTenCommand

    # test "command" processing strategy
    processor_as_command = MyProcessor(
        processing_strategy=CommandProcessingStrategy
    )

    result = await processor_as_command.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
    )
    assert result == 120

    # change the processing strategy to "processor"
    processor = processor_as_command.with_processing_strategy_class(
        ProcessorProcessingStrategy
    )

    result = await processor.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
    )
    assert result == 12


@pytest.mark.asyncio
async def test_command_with_extra_arguments():
    # define a command class
    class MultiplyCommand(Command[int, int]):
        async def execute(self, factor: int) -> int:
            for stage in self.stages:
                self.payload = await self._call_stage(
                    payload=self.payload,
                    stage=stage,
                )

            return self.payload * factor

    class MyProcessor(ChainedProcessor[int, int]):
        command_class = MultiplyCommand

    # test command
    processor = MyProcessor()

    result = await processor.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
        factor=8,
    )
    assert result == 96



from typing import Protocol

from ..partials.with_base.pipeline_interface import PipelineInterface
from ..partials.with_pipeline_factory.pipeline_factory_interface import (
    PipelineFactoryInterface as WithPipelineFactoryBaseInterface,
)
from ..partials.with_pipeline_processor.pipeline_factory_interface_mixin import (  # noqa
    PipelineFactoryInterfaceMixin as WithProcessorInterface,
)


class PipelineFactoryInterface(
    WithProcessorInterface,
    WithPipelineFactoryBaseInterface[PipelineInterface],
    Protocol,
):
    pass



from abc import abstractmethod
from typing import Any, Protocol, TypeVar

from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType


class ProcessorInterface(
    Protocol[T_in, T_out],
):
    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    async def _call_stage(
        self, payload: T_in, stage: StageCallableType, *args: Any, **kwds: Any
    ) -> T_in: ...


TProcessor = TypeVar(
    "TProcessor", bound=ProcessorInterface, infer_variance=True
)



from abc import ABC

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.stage_callable import StageCallableType
from ...support.act_as_list import HasListImmutability


class PipelineMixin(
    HasListImmutability[StageCallableType],
    ImplementsPipelineInterface,
    ABC,
):
    pass



from typing import Protocol

from .stage_facade import StageFacade as WithStageInterface
from ..with_base.types import T_in, T_out
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..with_pipeline_processor.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineProcessorInterface,
)


class PipelineInterfaceMixin(
    WithStageInterface[T_in, T_out],
    WithPipelineProcessorInterface[T_in, T_out],
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    pass



from typing import Any, Protocol, TypeVar


class PipelineInterface(
    Protocol,
):
    def __init__(self, *args: Any, **kwds: Any) -> None: ...


TPipeline = TypeVar("TPipeline", bound=PipelineInterface, infer_variance=True)



from abc import ABC, abstractmethod
from typing import Any

from .stage_interface_mixin import (
    StageInterfaceMixin as ImplementsStageInterface,
)
from ..with_base.types import T_in, T_out


class StageMixin(
    ImplementsStageInterface[T_in, T_out],
    ABC,
):
    @abstractmethod
    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        pass



from typing import Any, Generic, Optional, Self

from .pipeline_factory_interface import PipelineFactoryInterface
from ..with_base.pipeline_interface import TPipeline
from ..with_pipeline_declared_stages.pipeline_mixin import StageClassOrInstance
from ...support.act_as_factory import ActAsFactory
from ...support.act_as_list import ActAsList


class PipelineFactory(
    PipelineFactoryInterface[TPipeline],
    ActAsFactory[TPipeline],
    ActAsList[StageClassOrInstance],
    Generic[TPipeline],
):
    _instance_class: Optional[type[TPipeline]] = None  # ActAsFactory
    stages: list[StageClassOrInstance] = []  # ActAsList

    def __init__(
        self,
        stages: list[StageClassOrInstance] = [],
        *args: Any,
        **kwds: Any,
    ) -> None:
        if not self.stages:
            self.stages = [*stages] or []

    # ActAsList
    def _get_items(self) -> list[StageClassOrInstance]:
        return self.stages

    # ActAsList
    def _set_items(self, items: list[StageClassOrInstance]) -> Self:
        self.stages = items

        return self

    # ActAsFactory
    def _definition(self) -> dict[str, Any]:
        return {"stages": self._get_items()}

    # User factory configuration
    def add_stage(self, stage: StageClassOrInstance) -> Self:
        return self._add_item(stage)

    # User factory configuration
    def with_stages(self, stages: list[StageClassOrInstance]) -> Self:
        return self._set_items([*stages])



from typing import Protocol

from ..partials.with_base.types import T_in, T_out
from ..partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)
from ..partials.with_processor_commands.processor_interface_mixin import (
    ProcessorInterfaceMixin as WithProcessorCommandsInterface,
)


class ProcessorInterface(
    WithProcessorCommandsInterface[T_in, T_out],
    WithProcessorBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass



import copy
from abc import ABC
from typing import Any, Generic, Optional, Self

from .strategies.command_processing_strategy import CommandProcessingStrategy
from .strategies.processor_processing_strategy import (
    ProcessorProcessingStrategy,
)
from .processing_strategy_interface import ProcessingStrategyInterface
from .command_interface import CommandInterface
from .processor_interface_mixin import (
    ProcessorInterfaceMixin as ImplementsProcessorInterface,
)
from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType


class ProcessorMixin(
    ImplementsProcessorInterface[T_in, T_out],
    Generic[T_in, T_out],
    ABC,
):
    command_class: Optional[type[CommandInterface]] = None
    processing_strategy: Optional[type[ProcessingStrategyInterface]] = None
    processing_strategy_instance: ProcessingStrategyInterface

    def __init__(
        self,
        command_class: Optional[type[CommandInterface]] = None,
        processing_strategy: Optional[
            type[ProcessingStrategyInterface]
        ] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        self.command_class = command_class or self.command_class

        self.processing_strategy = (
            processing_strategy or self.which_processing_strategy_class()
        )

        self.processing_strategy_instance = self._make_processing_strategy()

    def with_processing_strategy(
        self, processing_strategy: ProcessingStrategyInterface[T_in, T_out]
    ) -> Self:
        cloned = copy.deepcopy(self)

        cloned.processing_strategy_instance = processing_strategy

        return cloned

    def with_processing_strategy_class(
        self,
        processing_strategy_class: type[
            ProcessingStrategyInterface[T_in, T_out]
        ],
    ) -> Self:
        cloned = copy.deepcopy(self)

        cloned.processing_strategy = processing_strategy_class

        cloned.processing_strategy_instance = processing_strategy_class(cloned)

        return cloned

    def which_processing_strategy_class(
        self,
    ) -> type[ProcessingStrategyInterface[T_in, T_out]]:
        if self.processing_strategy:
            return self.processing_strategy

        if self.command_class:
            return CommandProcessingStrategy

        return ProcessorProcessingStrategy

    def _make_processing_strategy(
        self,
    ) -> ProcessingStrategyInterface[T_in, T_out]:
        strategy = self.which_processing_strategy_class()

        return strategy(processor=self, command_class=self.command_class)

    async def process_with_strategy(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        return await self.processing_strategy_instance.process(
            payload, stages, *args, **kwds
        )

    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        raise NotImplementedError("Processor.process() must be implemented.")



from typing import Protocol, Self

from .item import TItem


class ActAsListInterface(
    Protocol[TItem],
):
    def _get_items(self) -> list[TItem]: ...

    def _set_items(self, items: list[TItem]) -> Self: ...

    def _add_item(self, item: TItem) -> Self: ...



from typing import Protocol

from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)


class PipelineInterfaceMixin(
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol,
):
    def _should_instantiate_stages(self) -> bool: ...

    def _instantiate_stages(self) -> None: ...



from abc import abstractmethod
from typing import Any, Optional

from .act_as_factory_interface import ActAsFactoryInterface, TModel


class ActAsFactory(ActAsFactoryInterface[TModel]):
    # set the class here or in `_get_instance_class`
    _instance_class: Optional[type[TModel]] = None

    @abstractmethod
    def _definition(self) -> dict[str, Any]:
        pass

    def _get_instance_class(self) -> type[TModel]:
        if self._instance_class is None:
            raise ValueError("Factory _instance_class not set")

        return self._instance_class

    def make(self) -> TModel:
        instance_class = self._get_instance_class()

        return instance_class(**self._definition())



from abc import ABC
from typing import Any, Generic, Optional, cast

from .processor_interface_mixin import (
    ProcessorInterfaceMixin as ProcessorInterface,
)
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.types import T_in, T_out


class PipelineMixin(
    ImplementsPipelineInterface[T_in, T_out],
    Generic[T_in, T_out],
    ABC,
):
    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out:
        self.processor = cast(Optional[ProcessorInterface], self.processor)

        if self.processor is None:
            raise ValueError("Processor not set")

        return await self.processor.process_with_strategy(
            payload=payload, stages=self._get_items(), *args, **kwds
        )



from abc import ABC, abstractmethod
import inspect
from typing import Any, Generic

from .processor_interface_mixin import (
    ProcessorInterfaceMixin as ProcessorInterface,
)
from .command_interface import CommandInterface
from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType


class Command(
    CommandInterface[T_in, T_out],
    Generic[T_in, T_out],
    ABC,
):
    processor: ProcessorInterface
    payload: T_in = None
    stages: list[StageCallableType] = []
    extra_args: Any = None
    extra_kwds: Any = None

    def __init__(
        self,
        processor: ProcessorInterface,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> None:
        self.processor = processor
        self.payload = payload
        self.stages = stages.copy()
        self.extra_args = args
        self.extra_kwds = kwds

    @abstractmethod
    async def execute(self, *args: Any, **kwds: Any) -> T_out:
        pass

    async def _call_stage(
        self, payload: T_in, stage: StageCallableType, *args: Any, **kwds: Any
    ) -> T_in:
        result = stage(payload, *args, **kwds)

        if inspect.isawaitable(result):
            return await result

        return result



from .classes.command import Command
from .classes.command_interface import CommandInterface
from .classes.pipeline_factory import PipelineFactory
from .classes.pipeline_factory_interface import PipelineFactoryInterface
from .classes.pipeline import Pipeline
from .classes.pipeline_interface import PipelineInterface
from .classes.processor import Processor
from .classes.processor_interface import ProcessorInterface
from .classes.stage import Stage
from .classes.stage_interface import StageInterface
from .classes.types import T_in, T_out
from .processors.chained_processor.chained_pipeline import ChainedPipeline
from .processors.chained_processor.chained_processor import ChainedProcessor
from .processors.interruptible_processor.interruptible_processor import (
    InterruptibleProcessor,
)
from .processors.interruptible_processor.interruptible_pipeline import (
    InterruptiblePipeline,
)
from .partials.with_base.stage_callable import StageCallable
from .partials.with_base.pipeline import Pipeline as WithPipelineBase
from .partials.with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from .partials.with_base.stage import Stage as WithStageBase
from .partials.with_base.stage_interface import (
    StageInterface as WithStageBaseInterface,
)
from .partials.with_pipeline_as_list.pipeline_mixin import (
    PipelineMixin as WithPipelineAsList,
)
from .partials.with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from .partials.with_pipeline_processor.pipeline_mixin import (
    PipelineMixin as WithPipelineProcessor,
)
from .partials.with_pipeline_processor.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineProcessorInterface,
)
from .partials.with_pipeline_processor.processor import (
    Processor as WithProcessorBase,
)
from .partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)
from .partials.with_pipeline_as_stage.pipeline_mixin import (
    PipelineMixin as WithPipelineAsStage,
)
from .partials.with_pipeline_as_stage.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsStageInterface,
)
from .partials.with_pipeline_as_immutable.pipeline_mixin import (
    PipelineMixin as WithPipelineAsImmutable,
)
from .partials.with_pipeline_as_immutable.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsImmutableInterface,
)
from .partials.with_stage_as_callable.stage_mixin import (
    StageMixin as WithStageAsCallable,
)
from .partials.with_stage_as_callable.stage_interface_mixin import (
    StageInterfaceMixin as WithStageAsCallableInterface,
)
from .partials.with_processor_commands.processing_strategy import (
    ProcessingStrategy,
)
from .partials.with_processor_commands.processing_strategy_interface import (
    ProcessingStrategyInterface,
)
from .partials.with_processor_commands.strategies.command_processing_strategy import (  # noqa
    CommandProcessingStrategy,
)
from .partials.with_processor_commands.strategies.processor_processing_strategy import (  # noqa
    ProcessorProcessingStrategy,
)
from .partials.with_processor_commands.traits.processable_as_command import (
    ProcessableAsCommand,
)


# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.18.0"

# Expose the public API
__all__ = [
    # Core
    "Command",
    "CommandInterface",
    "Pipeline",
    "PipelineInterface",
    "Stage",
    "StageInterface",
    "Processor",
    "ProcessorInterface",
    "PipelineFactory",
    "PipelineFactoryInterface",
    "ProcessingStrategy",
    "ProcessingStrategyInterface",
    # Traits
    "ProcessableAsCommand",
    # Types
    "T_in",
    "T_out",
    "StageCallable",
    # Processors
    "ChainedProcessor",
    "ChainedPipeline",
    "InterruptibleProcessor",
    "InterruptiblePipeline",
    # Strategies
    "CommandProcessingStrategy",
    "ProcessorProcessingStrategy",
    # Partials
    "WithPipelineBase",
    "WithPipelineBaseInterface",
    "WithStageBase",
    "WithStageBaseInterface",
    "WithPipelineAsList",
    "WithPipelineAsListInterface",
    "WithPipelineProcessor",
    "WithPipelineProcessorInterface",
    "WithProcessorBase",
    "WithProcessorBaseInterface",
    "WithPipelineAsStage",
    "WithPipelineAsStageInterface",
    "WithPipelineAsImmutable",
    "WithPipelineAsImmutableInterface",
    "WithStageAsCallable",
    "WithStageAsCallableInterface",
]



from typing import Any, Optional, Protocol, Self

from ..with_base.pipeline_interface import TPipeline
from ..with_pipeline_declared_stages.pipeline_mixin import StageClassOrInstance
from ...support.act_as_factory import ActAsFactoryInterface
from ...support.act_as_list import ActAsListInterface


class PipelineFactoryInterface(
    ActAsFactoryInterface[TPipeline],
    ActAsListInterface[StageClassOrInstance],
    Protocol[TPipeline],
):
    def __init__(
        self,
        stages: Optional[list[StageClassOrInstance]] = [],
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def add_stage(self, stage: StageClassOrInstance) -> Self: ...

    def with_stages(self, stages: list[StageClassOrInstance]) -> Self: ...



from abc import ABC
from typing import Any, Self

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.stage_callable import StageCallableType
from ...support.act_as_list import ActAsList


class PipelineMixin(
    ActAsList[StageCallableType],
    ImplementsPipelineInterface,
    ABC,
):
    stage_instances: list[StageCallableType]

    def __init__(
        self,
        stage_instances: list[StageCallableType] = [],
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        self.stage_instances = [*stage_instances]

    def _get_items(self) -> list[StageCallableType]:
        return self.stage_instances

    def _set_items(self, items: list[StageCallableType]) -> Self:
        self.stage_instances = items

        return self

    def pipe(self, stage: StageCallableType) -> Self:
        return self._add_item(stage)



from __future__ import annotations
from abc import abstractmethod
from typing import Any, Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from .processor_interface_mixin import (
        ProcessorInterfaceMixin as ProcessorInterface,
    )

from ..with_base.stage_callable import StageCallable
from ..with_base.types import T_in, T_out


class ProcessingStrategyInterface(Protocol[T_in, T_out]):
    def __init__(
        self, processor: ProcessorInterface, *args: Any, **kwds: Any
    ) -> None: ...

    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallable],
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .processor_mixin import ProcessorMixin as Processor

from .processing_strategy_interface import ProcessingStrategyInterface
from ..with_base.stage_callable import StageCallable
from ..with_base.types import T_in, T_out


class ProcessingStrategy(ProcessingStrategyInterface[T_in, T_out]):
    processor: Processor

    def __init__(self, processor: Processor, *args: Any, **kwds: Any) -> None:
        self.processor = processor

    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallable],
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



from abc import ABC, abstractmethod
import inspect
from typing import Any

from .processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType


class Processor(
    ImplementsProcessorInterface[T_in, T_out],
    ABC,
):
    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        pass

    async def _call_stage(
        self, payload: T_in, stage: StageCallableType, *args: Any, **kwds: Any
    ) -> T_in:
        result = stage(payload, *args, **kwds)

        if inspect.isawaitable(result):
            return await result

        return result



from typing import Protocol

from ..partials.with_base.types import T_in, T_out
from ..partials.with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..partials.with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..partials.with_pipeline_declared_stages.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineDeclaredStagesInterface,
)
from ..partials.with_pipeline_processor.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineProcessorInterface,
)
from ..partials.with_pipeline_as_stage.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsStageInterface,
)
from ..partials.with_pipeline_as_immutable.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsImmutableInterface,
)
from ..partials.with_processor_commands.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithProcessorCommandsInterface,
)


class PipelineInterface(
    WithPipelineAsImmutableInterface,
    WithPipelineAsStageInterface[T_in, T_out],
    WithProcessorCommandsInterface[T_in, T_out],
    WithPipelineProcessorInterface[T_in, T_out],
    WithPipelineDeclaredStagesInterface,
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    pass



from typing import Protocol

from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)


class PipelineInterfaceMixin(
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol,
):
    pass



import pytest
import asyncio
from typing import AsyncIterator

from thecodecrate_pipeline import Pipeline


# Define stages
async def stage1(stream: AsyncIterator[int]) -> AsyncIterator[int]:
    async for item in stream:
        # Multiply each item by 2
        yield item * 2
        await asyncio.sleep(0.01)  # Simulate processing delay


async def stage2(stream: AsyncIterator[int]) -> AsyncIterator[str]:
    async for item in stream:
        # Convert each item to a formatted string
        yield f"Number: {item}"
        await asyncio.sleep(0.01)


# Create the pipeline
pipeline = (
    Pipeline[AsyncIterator[int], AsyncIterator[str]]()
    .pipe(stage1)
    .pipe(stage2)
)


# Define the input stream
async def input_stream() -> AsyncIterator[int]:
    for i in range(5):
        yield i
        await asyncio.sleep(0.01)


@pytest.mark.asyncio
async def test_pipeline_streams():
    # Process the input stream through the pipeline
    stream = await pipeline.process(input_stream())

    # Collect the results
    results: list[str] = []
    async for result in stream:
        results.append(result)

    # Assert that the results are as expected
    assert results == [
        "Number: 0",
        "Number: 2",
        "Number: 4",
        "Number: 6",
        "Number: 8",
    ]



from __future__ import annotations
from typing import Any, Optional, Protocol, Self
import typing

if typing.TYPE_CHECKING:
    from .processing_strategy_interface import ProcessingStrategyInterface
    from .command_interface import CommandInterface

from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType
from ..with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)


class ProcessorInterfaceMixin(
    WithProcessorBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        command_class: Optional[type[CommandInterface]] = None,
        processing_strategy: Optional[
            type[ProcessingStrategyInterface]
        ] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def with_processing_strategy(
        self, processing_strategy: ProcessingStrategyInterface[T_in, T_out]
    ) -> Self: ...

    def which_processing_strategy_class(
        self,
    ) -> type[ProcessingStrategyInterface[T_in, T_out]]: ...

    def _make_processing_strategy(
        self,
    ) -> ProcessingStrategyInterface[T_in, T_out]: ...

    async def process_with_strategy(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    # make `process` method not abstract
    # when using command strategy, the `process` method is not used
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



import pytest
from thecodecrate_pipeline import (
    Pipeline,
    ChainedPipeline,
    ChainedProcessor,
    InterruptiblePipeline,
    InterruptibleProcessor,
)

from tests.stubs.stub_processor import StubProcessor


@pytest.mark.asyncio
async def test_custom_processor():
    processor = StubProcessor()

    pipeline = (
        (Pipeline[int](processor=processor))
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload + 1)
    )

    assert await pipeline.process(1) == 30


@pytest.mark.asyncio
async def test_chained_pipeline():
    pipeline = (
        ChainedPipeline[int]()
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
    )
    assert await pipeline.process(5) == 12


@pytest.mark.asyncio
async def test_chained_processor():
    processor = ChainedProcessor[int]()

    result = await processor.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
    )
    assert result == 12

    result = await processor.process_with_strategy(
        payload=5,
        stages=[
            lambda payload: payload + 1,
        ],
    )
    assert result == 6


@pytest.mark.asyncio
async def test_pipeline_with_chained_processor():
    processor = ChainedProcessor[int]()

    pipeline = (
        (Pipeline[int](processor=processor))
        .pipe(lambda payload: payload + 2)
        .pipe(lambda payload: payload * 5)
        .pipe(lambda payload: payload * 10)
    )

    assert await pipeline.process(5) == 350
    assert await pipeline.process(1) == 150


@pytest.mark.asyncio
async def test_interruptible_pipeline():
    pipeline = (
        InterruptiblePipeline[int](lambda payload: payload < 10)
        .pipe(lambda payload: payload + 2)
        .pipe(lambda payload: payload * 10)
        .pipe(lambda payload: payload * 10)
    )
    assert await pipeline.process(5) == 70


@pytest.mark.asyncio
async def test_interruptible_processor():
    processor = InterruptibleProcessor[int](lambda payload: payload < 10)

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 2,
            lambda payload: payload * 10,
            lambda payload: payload * 10,
        ],
    )
    assert result == 70

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 2,
        ],
    )
    assert result == 7


@pytest.mark.asyncio
async def test_pipeline_with_interruptible_processor():
    processor = InterruptibleProcessor[int](lambda payload: payload < 20)

    pipeline = (
        (Pipeline[int](processor=processor))
        .pipe(lambda payload: payload + 2)
        .pipe(lambda payload: payload * 5)
        .pipe(lambda payload: payload * 10)
    )

    assert await pipeline.process(5) == 35
    assert await pipeline.process(1) == 150



from typing import Any

from ...with_base.stage_callable import StageCallable
from ...with_base.types import T_in, T_out
from ..processing_strategy import ProcessingStrategy


class ProcessorProcessingStrategy(ProcessingStrategy[T_in, T_out]):
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallable],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        return await self.processor.process(
            payload=payload, stages=stages, *args, **kwds
        )



from abc import abstractmethod
from typing import Any, Protocol, TypeVar

TModel = TypeVar("TModel", infer_variance=True)


class ActAsFactoryInterface(Protocol[TModel]):
    @abstractmethod
    def _definition(self) -> dict[str, Any]: ...

    def _get_instance_class(self) -> type[TModel]: ...

    def make(self) -> Any: ...



from abc import ABC

from .processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..partials.with_base.types import T_in, T_out
from ..partials.with_pipeline_processor.processor import (
    Processor as WithProcessorBaseConcern,
)
from ..partials.with_processor_commands.processor_mixin import (
    ProcessorMixin as WithProcessorCommandsConcern,
)


class Processor(
    WithProcessorCommandsConcern[T_in, T_out],
    WithProcessorBaseConcern[T_in, T_out],
    ImplementsProcessorInterface[T_in, T_out],
    ABC,
):
    pass



from typing import Protocol, Self

from .processor_interface import ProcessorInterface
from ..with_base.types import T_in, T_out
from ..with_base.pipeline_interface import PipelineInterface
from ..with_pipeline_factory.pipeline_factory_interface import (
    PipelineFactoryInterface as WithPipelineFactoryBaseInterface,
)


class PipelineFactoryInterfaceMixin(
    WithPipelineFactoryBaseInterface[PipelineInterface],
    Protocol[T_in, T_out],
):
    def with_processor(self, processor: ProcessorInterface) -> Self: ...

    def with_processor_class(
        self, processor_class: type[ProcessorInterface]
    ) -> Self: ...



from typing import Any

from .stage_facade import StageFacade
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.stage_callable import StageCallableType

StageClassOrInstance = type[StageFacade[Any, Any]] | StageCallableType


class PipelineMixin(
    ImplementsPipelineInterface,
):
    stages: list[StageClassOrInstance] = []

    def __init__(
        self, stages: list[StageClassOrInstance] = [], *args: Any, **kwds: Any
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if len(stages) > 0:
            self.stages = [*stages]

        if self._should_instantiate_stages():
            self._instantiate_stages()

    def _should_instantiate_stages(self) -> bool:
        return len(self._get_items()) == 0 and len(self.stages) > 0

    def _instantiate_stages(self) -> None:
        instances = [
            stage() if isinstance(stage, type) else stage
            for stage in self.stages
        ]

        self._set_items(instances)



from abc import ABC

from .command_interface import (
    CommandInterface as ImplementsCommandInterface,
)
from ..partials.with_base.types import T_in, T_out
from ..partials.with_processor_commands.command import (
    Command as WithCommandBaseConcern,
)


class Command(
    WithCommandBaseConcern[T_in, T_out],
    ImplementsCommandInterface[T_in, T_out],
    ABC,
):
    pass



from .act_as_list import ActAsList
from .act_as_list_interface import ActAsListInterface
from .item import TItem
from .traits.has_list_immutability import HasListImmutability

__all__ = [
    # Core
    "ActAsList",
    "ActAsListInterface",
    "TItem",
    # Traits
    "HasListImmutability",
]



from typing import Any, Protocol, TypeVar

from ..partials.with_base.types import T_in, T_out
from ..partials.with_processor_commands.command_interface import (
    CommandInterface as WithCommandBaseInterface,
)


class CommandInterface(
    WithCommandBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass


TCommand = TypeVar(
    "TCommand", bound=CommandInterface[Any, Any], infer_variance=True
)



import pytest
from thecodecrate_pipeline import (
    Pipeline,
    StageCallable,
)

from tests.stubs.stub_stage_with_custom_args import (
    IndexedPipeline,
    IndexedStage,
)
from .stubs.stub_stages_int import (
    AddOneStage,
    TimesThreeStage,
    TimesTwoStage,
)
from .stubs.stub_stage import StubStage
from .stubs.stub_processor import StubProcessor


@pytest.mark.asyncio
async def test_lambda_stages():
    pipeline = (
        (Pipeline[int]())
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
        .pipe(lambda payload: payload + 1)
    )

    assert await pipeline.process(1) == 5
    assert await pipeline.process(2) == 7


@pytest.mark.asyncio
async def test_classbased_stages():
    pipeline = (
        (Pipeline[int]())
        .pipe(TimesTwoStage())
        .pipe(AddOneStage())
        .pipe(TimesThreeStage())
    )

    assert await pipeline.process(5) == 33


@pytest.mark.asyncio
async def test_pipelinebased_stages():
    sub_pipeline1 = (
        (Pipeline[int]())
        .pipe(TimesTwoStage())
        .pipe(AddOneStage())
        .pipe(TimesThreeStage())
    )

    sub_pipeline2 = (
        (Pipeline[int]())
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
    )

    pipeline = (
        (Pipeline[int]())
        .pipe(TimesThreeStage())
        .pipe(sub_pipeline1)
        .pipe(sub_pipeline2)
        .pipe(lambda payload: payload * 2)
    )

    assert await pipeline.process(10) == 736


@pytest.mark.asyncio
async def test_string_stage():
    pipeline = (Pipeline[str]()).pipe(StubStage())

    assert await pipeline.process("") == "stubbed response"


@pytest.mark.asyncio
async def test_process_repeat():
    # it happened that the pipeline was not reset, so the second process call
    # would not work as expected
    for _ in range(3):
        pipeline = (
            (Pipeline[int]())
            .pipe(lambda payload: payload + 1)
            .pipe(lambda payload: payload * 2)
            .pipe(lambda payload: payload + 1)
        )

        for __ in range(3):
            assert await pipeline.process(1) == 5
            assert await pipeline.process(2) == 7


@pytest.mark.asyncio
async def test_pipeline_immutability():
    pipeline = (
        (Pipeline[int]())
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload + 1)
    )

    assert await pipeline.process(1) == 3

    pipeline2 = pipeline.pipe(lambda payload: payload + 1)

    assert await pipeline.process(1) == 3
    assert await pipeline2.process(1) == 4


@pytest.mark.asyncio
async def test_declarative_stages():
    pipeline = (Pipeline[int]()).pipe(TimesTwoStage()).pipe(AddOneStage())

    def add_seven(payload: int) -> int:
        return payload + 7

    async def sub_three_async(payload: int) -> int:
        return payload - 3

    class MyPipeline(Pipeline[int]):
        stages = [
            TimesThreeStage(),  # stage instance
            pipeline,  # pipeline
            AddOneStage,  # stage class
            add_seven,  # function
            sub_three_async,  # async function
        ]

    assert await MyPipeline().process(5) == 36


@pytest.mark.asyncio
async def test_declarative_stages_with_processor():
    class MyPipeline(Pipeline[int]):
        processor_class = StubProcessor
        stages = [
            TimesTwoStage,
            TimesThreeStage,
        ]

    assert await MyPipeline().process(5) == 300


@pytest.mark.asyncio
async def test_declarative_stages_dont_override_constructor_stages():
    class MyPipeline(Pipeline[int]):
        stages = [
            TimesTwoStage,
            TimesThreeStage,
        ]

    stage_instances: list[StageCallable] = [AddOneStage(), TimesTwoStage()]

    assert await MyPipeline(stage_instances=stage_instances).process(5) == 12


@pytest.mark.asyncio
async def test_stages_with_custom_args():
    class MyIndexedStage(IndexedStage[str]):
        async def __call__(self, payload: str, index: int) -> str:
            return f"{payload}: {index}"

    pipeline = (
        (IndexedPipeline[str]()).pipe(MyIndexedStage()).pipe(MyIndexedStage())
    )

    assert await pipeline.process("test") == "test: 0: 1"



from typing import Protocol, Self

from ..with_base.stage_callable import StageCallableType
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ...support.act_as_list.act_as_list_interface import (
    ActAsListInterface,
)


class PipelineInterfaceMixin(
    ActAsListInterface[StageCallableType],
    WithPipelineBaseInterface,
    Protocol,
):
    def pipe(self, stage: StageCallableType) -> Self: ...



from abc import ABC
from typing import Any, Optional

from .processor_interface import ProcessorInterface
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as PipelineInterface,
)
from ..with_base.types import T_in, T_out


class PipelineMixin(
    PipelineInterface[T_in, T_out],
    ABC,
):
    processor_class: Optional[type[ProcessorInterface]] = None
    processor: Optional[ProcessorInterface] = None

    def __init__(
        self,
        processor_class: Optional[type[ProcessorInterface]] = None,
        processor: Optional[ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if processor_class:
            self.processor_class = processor_class

        if processor:
            self.processor = processor

        if not self.processor:
            self.processor = self._make_processor()

    def _make_processor(self) -> ProcessorInterface:
        if self.processor_class is None:
            raise ValueError("Processor class not set")

        return self.processor_class()

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out:
        if self.processor is None:
            raise ValueError("Processor not set")

        return await self.processor.process(
            payload=payload, stages=self._get_items(), *args, **kwds
        )



from typing import Any
from .pipeline_interface import (
    PipelineInterface as ImplementsPipelineInterface,
)


class Pipeline(
    ImplementsPipelineInterface,
):
    def __init__(self, *args: Any, **kwds: Any) -> None:
        pass



from typing import Any, Optional, cast

from ...with_base.types import T_in, T_out
from ..processor_mixin import ProcessorMixin as Processor
from ..strategies.command_processing_strategy import CommandProcessingStrategy
from ..pipeline_interface_mixin import PipelineInterfaceMixin


class ProcessableAsCommand(PipelineInterfaceMixin[T_in, T_out]):
    def __init__(self, *args: Any, **kwds: Any) -> None:
        self.processor = cast(
            Optional[Processor], self.processor
        )  # type: ignore

        super().__init__(*args, **kwds)  # type: ignore

        if self.processor:
            self.processor.with_processing_strategy(
                CommandProcessingStrategy(
                    self.processor, self.processor.command_class
                )
            )



from typing import Any, Protocol

from ..with_base.types import T_in, T_out
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..with_pipeline_processor.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineProcessorInterface,
)


class PipelineInterfaceMixin(
    WithPipelineProcessorInterface[T_in, T_out],
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    async def process(
        self, payload: T_in, *args: Any, **kwds: Any
    ) -> T_out: ...



from typing import Protocol

from ..with_base.types import T_in, T_out
from ..with_base.stage_interface import (
    StageInterface as WithStageBaseInterface,
)
from ..with_stage_as_callable.stage_interface_mixin import (
    StageInterfaceMixin as WithStageAsCallableInterface,
)


class StageFacade(
    WithStageAsCallableInterface[T_in, T_out],
    WithStageBaseInterface,
    Protocol[T_in, T_out],
):
    pass



from typing import Generic

from .pipeline import Pipeline
from ..partials.with_base.types import T_in, T_out
from ..partials.with_pipeline_factory.pipeline_factory import (
    PipelineFactory as WithPipelineFactoryBaseConcern,
)
from .pipeline_factory_interface import (
    PipelineFactoryInterface as ImplementsPipelineFactoryInterface,
)
from ..partials.with_pipeline_processor.pipeline_factory_mixin import (
    PipelineFactoryMixin as WithProcessorConcern,
)


class PipelineFactory(
    WithProcessorConcern,
    WithPipelineFactoryBaseConcern[Pipeline[T_in, T_out]],
    ImplementsPipelineFactoryInterface,
    Generic[T_in, T_out],
):
    _instance_class = Pipeline



import inspect
from typing import Any, Awaitable, Callable, cast

from ...classes.stage_callable import StageCallable
from ...classes.types import T_in, T_out
from ...classes.processor import Processor

CheckCallable = Callable[[T_in], bool | Awaitable[bool]]


class InterruptibleProcessor(Processor[T_in, T_out]):
    check: CheckCallable[T_in]

    def __init__(self, check: CheckCallable[T_in]) -> None:
        super().__init__()

        self.check = check

    async def process(
        self,
        payload: T_in,
        stages: list[StageCallable],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        for stage in stages:
            payload = await self._call_stage(
                payload=payload, stage=stage, *args, **kwds
            )

            if not await self._call_check(payload):
                return cast(T_out, payload)

        return cast(T_out, payload)

    async def _call_check(self, payload: T_in) -> bool:
        result = self.check(payload)

        if inspect.isawaitable(result):
            return await result

        return result



from typing import Any

from ...with_base.stage_callable import StageCallable
from ...with_base.types import T_in, T_out
from ..processing_strategy import ProcessingStrategy


class CommandProcessingStrategy(ProcessingStrategy[T_in, T_out]):
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallable],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        if not self.processor.command_class:
            raise ValueError("Command class is not set")

        command = self.processor.command_class(
            processor=self.processor,
            payload=payload,
            stages=stages,
            *args,
            **kwds,
        )

        return await command.execute(*args, **kwds)



from typing import Any, cast

from ...classes.stage_callable import StageCallable
from ...classes.types import T_in, T_out
from ...classes.processor import Processor
from .chained_command import ChainedCommand


class ChainedProcessor(Processor[T_in, T_out]):
    command_class = ChainedCommand

    async def process(
        self,
        payload: T_in,
        stages: list[StageCallable],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        for stage in stages:
            payload = await self._call_stage(
                payload=payload, stage=stage, *args, **kwds
            )

        return cast(T_out, payload)
